package org.jastadd.ragconnect.compiler;

import beaver.Parser;
import org.jastadd.option.BooleanOption;
import org.jastadd.option.ValueOption;
import org.jastadd.relast.compiler.AbstractCompiler;
import org.jastadd.relast.compiler.CompilerException;
import org.jastadd.ragconnect.ast.*;
import org.jastadd.ragconnect.parser.RagConnectParser;
import org.jastadd.ragconnect.scanner.RagConnectScanner;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.Collection;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

public class Compiler extends AbstractCompiler {

//  private ValueOption optionOutputDir;
  private ValueOption optionRootNode;
  private BooleanOption optionVerbose;
  private BooleanOption optionLogReads;
  private BooleanOption optionLogWrites;

  public Compiler() {
    super("ragconnect", true);
  }

  @Override
  protected int compile() throws CompilerException {
    if (getConfiguration().shouldPrintVersion()) {
      System.out.println(readVersion());
      return 0;
    }
    if (getConfiguration().shouldPrintHelp()) {
      getConfiguration().printHelp(System.out);
      return 0;
    }

    printMessage("Running RagConnect " + readVersion());

    if (!getConfiguration().outputDir().exists()) {
      try {
        Files.createDirectories(getConfiguration().outputDir().toPath());
      } catch (IOException e) {
        throw new CompilerException("Error creating output dir " + getConfiguration().outputDir(), e);
      }
    }

    RagConnect ragConnect = parseProgram(getConfiguration().getFiles());

    if (!ragConnect.errors().isEmpty()) {
      System.err.println("Errors:");
      for (ErrorMessage e : ragConnect.errors()) {
        System.err.println(e);
      }
      System.exit(1);
    }

    printMessage("Writing output files");
    // copy MqttHandler into outputDir
    final String mqttHandlerFileName = "MqttHandler.jadd";
    try {
      InputStream inputStream = Compiler.class.getClassLoader().getResourceAsStream(mqttHandlerFileName);
      if (inputStream == null) {
        throw new CompilerException("Could not open " + mqttHandlerFileName);
      }
      Files.copy(inputStream,
          getConfiguration().outputDir().toPath().resolve(mqttHandlerFileName),
          StandardCopyOption.REPLACE_EXISTING);
    } catch (IOException e) {
      throw new CompilerException("Could not copy " + mqttHandlerFileName, e);
    }
    for (GrammarFile grammarFile : ragConnect.getProgram().getGrammarFileList()) {
      Path outputFile = getConfiguration().outputDir().toPath().resolve(grammarFile.getFileName());
      writeToFile(outputFile, grammarFile.generateAbstractGrammar());
    }
    writeToFile(getConfiguration().outputDir().toPath().resolve("RagConnect.jadd"), ragConnect.generateAspect(optionRootNode.value()));
    return 0;
  }

  public static void main(String[] args) {
    System.setProperty("java.util.logging.manager", "org.apache.logging.log4j.jul.LogManager");
    System.setProperty("mustache.debug", "true");
    try {
      new Compiler().run(args);
    } catch (CompilerException e) {
      System.err.println(e.getMessage());
      System.exit(-1);
    }
  }

  /**
   * Reads the version string.
   *
   * The version string is read from the property file
   * src/main/resources/Version.properties. This
   * file should be generated during the build process. If it is missing
   * then there is some problem in the build script.
   *
   * @author Jesper Öqvist <jesper.oqvist@cs.lth.se>
   * @return the read version string, or <code>version ?</code>
   */
  private String readVersion() {
    try {
      ResourceBundle resources = ResourceBundle.getBundle("ragConnectVersion");
      return resources.getString("version");
    } catch (MissingResourceException e) {
      return "version ?";
    }
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(Path path, String str) throws CompilerException {
    try (BufferedWriter writer = Files.newBufferedWriter(path)) {
      writer.append(str);
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + path.toAbsolutePath(), e);
    }
  }

  protected void initOptions() {
    super.initOptions();
    optionRootNode = addOption(
        new ValueOption("rootNode", "root node in the base grammar.")
        .acceptAnyValue()
        .needsValue(true));
    optionVerbose = addOption(
        new BooleanOption("verbose", "Print more messages while compiling.")
        .defaultValue(false));
    optionLogReads = addOption(
        new BooleanOption("logReads", "Enable logging for every read.")
        .defaultValue(false));
    optionLogWrites = addOption(
        new BooleanOption("logWrites", "Enable logging for every write.")
        .defaultValue(false));
  }

  private RagConnect parseProgram(Collection<String> files) throws CompilerException {
    Program program = new Program();

    RagConnect ragConnect = new RagConnect();
    ragConnect.setProgram(program);

    GrammarFile ragConnectGrammarPart = new GrammarFile();
    ragConnectGrammarPart.setFileName("RagConnect.relast");
    program.addGrammarFile(ragConnectGrammarPart);

    for (String filename : files) {
      String extension = filename.substring(filename.lastIndexOf('.') + 1);
      switch (extension) {
        case "ast":
        case "relast":
          // processGrammar
          parseGrammar(program, filename);
          break;
        case "connect":
          // process ragConnect
          RagConnect parsedRagConnect = parseRagConnect(program, filename);
          mergeRagConnectDefinitions(ragConnect, parsedRagConnect);
          break;
        default:
          throw new CompilerException("Unknown file extension in " + filename);
      }
    }
    ragConnect.treeResolveAll();
    ragConnect.additionalRelations().forEach(ragConnectGrammarPart::addDeclaration);
    ASTNode.loggingEnabledForReads = optionLogReads.value();
    ASTNode.loggingEnabledForWrites = optionLogWrites.value();
    return ragConnect;
  }

  private void parseGrammar(Program program, String filename) throws CompilerException {
    try (BufferedReader reader = Files.newBufferedReader(Paths.get(filename))) {
      RagConnectScanner scanner = new RagConnectScanner(reader);
      RagConnectParser parser = new RagConnectParser();
      GrammarFile grammarFile = (GrammarFile) parser.parse(scanner);
      if (optionVerbose.value()) {
        grammarFile.dumpTree(System.out);
      }
      program.addGrammarFile(grammarFile);
      grammarFile.treeResolveAll();
      grammarFile.setFileName(toBaseName(filename));
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse grammar file " + filename, e);
    }
  }

  private RagConnect parseRagConnect(Program program, String filename) throws CompilerException {
    try (BufferedReader reader = Files.newBufferedReader(Paths.get(filename))) {
      RagConnectScanner scanner = new RagConnectScanner(reader);
      RagConnectParser parser = new RagConnectParser();
      RagConnect ragConnect = (RagConnect) parser.parse(scanner, RagConnectParser.AltGoals.ragconnect);
      ragConnect.setProgram(program);
      ragConnect.setFileName(toBaseName(filename));
      return ragConnect;
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse connect file " + filename, e);
    }
  }

  /**
   * Extracts the basename of the given file, with file extension
   * @param filename the given filename
   * @return the basename
   */
  private String toBaseName(String filename) {
    return new File(filename).getName();
  }

  private void mergeRagConnectDefinitions(RagConnect ragConnect, RagConnect ragConnectToIntegrate) {
    for (EndpointDefinition endpointDefinition : ragConnectToIntegrate.getEndpointDefinitionList()) {
      ragConnect.addEndpointDefinition(endpointDefinition);
    }
    for (MappingDefinition mappingDefinition : ragConnectToIntegrate.getMappingDefinitionList()) {
      ragConnect.addMappingDefinition(mappingDefinition);
    }
    for (DependencyDefinition dependencyDefinition : ragConnectToIntegrate.getDependencyDefinitionList()) {
      ragConnect.addDependencyDefinition(dependencyDefinition);
    }
  }

//  protected void printUsage() {
//    System.out.println("Usage: java -jar ragconnect.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
//    System.out.println("Options:");
//    System.out.print(commandLine.printOptionHelp());
//  }
}
